/**
 * Copyright 2006-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.xmlconfig;

import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.xml.DocumentParserException;
import org.ow2.util.xmlconfig.mapping.AttributeMapping;
import org.ow2.util.xmlconfig.mapping.ClassMapping;
import org.ow2.util.xmlconfig.mapping.XMLMapping;
import org.ow2.util.xmlconfig.mapping.XMLMappingBuilder;
import org.ow2.util.xmlconfig.properties.IPropertyResolver;
import org.ow2.util.xmlconfig.properties.PropertyResolverHelper;
import org.ow2.util.xmlconfig.source.IConfigurationSource;
import org.ow2.util.xmlconfig.source.InputStreamConfigurationSource;
import org.ow2.util.xmlconfig.source.XMLConfigurationSource;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * XML configuration class that configure a given object with an xml
 * configuration file. <br>
 * Additional information will be searched on the namespace with the resource named easybeans-mapping.xml.
 * @author Florent Benoit
 */
public class XMLConfiguration {

    /**
     * Http:// namespace.
     */
    private static final String HTTP_NAMESPACE = "http://";

    /**
     * By default, no validation.
     */
    private static final boolean VALIDATING_OFF = false;

    /**
     * Per-Package mapping file name.
     */
    private final String mappingFilename;

    /**
     * Mapping suffix.
     */
    private static final String MAPPING_SUFFIX = "-mapping.xml";

    /**
     * Minimum of property length: Based on ${P} with P the name of the property.
     */
    private static final int MIN_PROPERTY_LENGTH = 4;

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(XMLConfiguration.class);

    /**
     * Validating when parsing xml ?
     */
    private boolean validating = VALIDATING_OFF;

    /**
     * Mapping between namespace and XMLMapping object.
     */
    private Map<String, XMLMapping> mappings = null;

    /**
     * Link between name and the objects.
     */
    private HashMap<String, Object> configuredElements = null;

    /**
     * Link between name and the objects.
     */
    private Map<String, Object> contextualInstances = null;

    /**
     * Property resolver.
     */
    private IPropertyResolver propertyResolver = null;

    /**
     * List of the sources of configuration to be processed.
     */
    private List<IConfigurationSource> sources;

    /**
     * Build an xml configuration with the given xml configuration file.
     * @param xmlConfigurationURL the given xml configuration file
     * @param mappingFilename per-Package mapping file name
     * @param mappingNS the namespace of the mapping
     * @deprecated
     */
    @Deprecated
    public XMLConfiguration(final URL xmlConfigurationURL, final String mappingNS, final String mappingFilename) {
        this(xmlConfigurationURL, mappingFilename);
    }

    /**
     * Build an xml configuration with the given xml configuration file.
     * @param xmlConfigurationURL the given xml configuration file
     * @param mappingFilename per-Package mapping file name
     */
    public XMLConfiguration(final URL xmlConfigurationURL, final String mappingFilename) {
        this(mappingFilename);
        sources.add(new XMLConfigurationSource(xmlConfigurationURL));
    }

    /**
     * Build an xml configuration with the given xml configuration file.
     * @param stream the given xml configuration file
     * @param mappingFilename per-Package mapping file name
     */
    public XMLConfiguration(final InputStream stream, final String mappingFilename) {
        this(mappingFilename);
        sources.add(new InputStreamConfigurationSource("Unamed stream", stream));
    }

    /**
     * Initialize class fields.
     * @param mappingFilename per-Package mapping file name
     */
    public XMLConfiguration(final String mappingFilename) {
        this.mappings = new HashMap<String, XMLMapping>();
        this.configuredElements = new HashMap<String, Object>();
        this.contextualInstances = new HashMap<String, Object>();
        this.sources = new ArrayList<IConfigurationSource>();
        this.mappingFilename = mappingFilename;
    }

    /**
     * Add an URL pointing to an XML configuration file that will be used
     * sequentially to configure an instance.
     * @param xmlConfigurationFile URL pointing to an XML configuration file
     */
    public void addConfigurationFile(final URL xmlConfigurationFile) {
        sources.add(new XMLConfigurationSource(xmlConfigurationFile));
    }

    /**
     * Add an {@link InputStream} that will be used sequentially to configure an instance.
     * @param name Stream's name
     * @param stream the {@link InputStream} wrapping an XML configuration
     */
    public void addConfigurationStream(final String name, final InputStream stream) {
        sources.add(new InputStreamConfigurationSource(name, stream));
    }

    /**
     * Add a new {@link IConfigurationSource} wrapping an XML configuration content.
     * @param source the configuration's source
     */
    public void addConfigurationSource(final IConfigurationSource source) {
        sources.add(source);
    }

    /**
     * Configure the given object with the configuration that was analyzed.
     * @param object the object to configure.
     * @throws XMLConfigurationException if configuration fails
     */
    public void configure(final Object object) throws XMLConfigurationException {

        for (IConfigurationSource source : sources) {
            Element rootElement = analyzeXmlFile(source);
            setAttributesOfRootElement(object, rootElement);
            configure(object, rootElement, null);
        }
    }

    /**
     * Set the attributes of the root element.
     * @param object the given object to configure.
     * @param element the element that contains configuration.
     * @throws XMLConfigurationException if configuration fails
     */
    private void setAttributesOfRootElement(final Object object, final Element element) throws XMLConfigurationException {
        ClassMapping classMapping;
        try {
            classMapping = getClassMapping(element.getNamespaceURI(), element.getLocalName());
        } catch (XMLConfigurationException e) {
            // Only catch the exception because use of attributes on root element is optional
            logger.debug("No mapping for the root element.");
            return;
        }
        // Set attributes
        setAttributes(element, object, classMapping);
    }

    /**
     * Gets the root element of the XML parsed file.
     * @return the element object
     * @throws XMLConfigurationException if analyze fails.
     */
    private Element analyzeXmlFile(final IConfigurationSource source) throws XMLConfigurationException {
        // Parse the XML file and build all configuration process.
        Document xmlConfigurationDocument = null;
        try {
            xmlConfigurationDocument = source.getDocument(validating);
        } catch (DocumentParserException e) {
            throw new XMLConfigurationException("Cannot get a document on the given url '" + source.getName() + "'.", e);
        }

        // Get the root element
        Element rootElement = xmlConfigurationDocument.getDocumentElement();

        // get namespace of the element
        String nameSpace = rootElement.getNamespaceURI();

        // init mapping for the element
        getXMLMapping(nameSpace, rootElement.getLocalName());

        return rootElement;

    }


    /**
     * Is that the given name should be considered as an XML element (and read the value of this element).
     * @param name the given name
     * @param classMapping the mapping used to get the info
     * @return true if the name should be read from the XML element.
     */
    public boolean isElementMapping(final String name, final ClassMapping classMapping) {
        AttributeMapping attributeMapping = classMapping.getAttributeMapping(name);
        if (attributeMapping != null) {
            return attributeMapping.isElement();
        }
        return false;
    }

    /**
     * Is that the given name should be used with a list.
     * @param name the given name
     * @param classMapping the mapping used to get the info
     * @return true if the name should be used as a list.
     */
    public boolean isListElement(final String name, final ClassMapping classMapping) {
        AttributeMapping attributeMapping = classMapping.getAttributeMapping(name);
        if (attributeMapping != null) {
            return attributeMapping.isListElement();
        }
        return false;
    }

    /**
     * Configure the given object by looking at the element (of the
     * configuration).
     * @param object the given object to configure.
     * @param element the element that contains configuration.
     * @param parentClassMapping the parent class mapping
     * @throws XMLConfigurationException if configuration fails
     */
    public void configure(final Object object, final Element element, final ClassMapping parentClassMapping)
            throws XMLConfigurationException {

        // Get all children of the given element
        NodeList nodeList = element.getChildNodes();

        // For each element, create object and configure it
        int length = nodeList.getLength();
        for (int i = 0; i < length; i++) {
            Node node = nodeList.item(i);

            // Get an element, create an instance of the element
            if (Node.ELEMENT_NODE == node.getNodeType()) {

                // Is it an element that need to be considered as a string value
                if (parentClassMapping != null && isElementMapping(node.getLocalName(), parentClassMapping)) {
                    // Get value
                    Node txtNode = node.getFirstChild();
                    String val = null;
                    if (txtNode != null) {
                        val = txtNode.getNodeValue();
                    }

                    // Do nothing if value is null
                    if (val == null) {
                        continue;
                    } else {
                        // cleanup the value by removing extra spaces
                        val = val.trim();
                    }

                    // Call setter directly
                    if (!isListElement(node.getLocalName(), parentClassMapping)) {
                        setAttribute(object, node.getLocalName(), val, parentClassMapping);
                    } else {
                        // Need to call the getters and setters
                        addSetElement(object, val, node.getLocalName(), parentClassMapping);
                    }
                    continue;
                }
                ClassMapping classMapping = getClassMapping(node.getNamespaceURI(), node.getLocalName());

                // Build an object for the given class mapping
                Object newObject = newInstance(node, classMapping, object.getClass().getClassLoader());

                // Try to see if there is a value for the current object
                Method getterMethod = findSingleGetterMethod(object, newObject);
                if (getterMethod != null) {
                    // Call the getter method
                    logger.debug("Calling method {0} on object {1}", getterMethod, object);
                    Object getterObject = null;
                    try {
                        getterObject = getterMethod.invoke(object);
                    } catch (IllegalArgumentException e) {
                        throw new XMLConfigurationException("Cannot call method '" + getterMethod + "' on object '" + object
                                + "'.", e);
                    } catch (IllegalAccessException e) {
                        throw new XMLConfigurationException("Cannot call method '" + getterMethod + "' on object '" + object
                                + "'.", e);
                    } catch (InvocationTargetException e) {
                        throw new XMLConfigurationException("Cannot call method '" + getterMethod + "' on object '" + object
                                + "'.", e);
                    }

                    // find an object ?
                    if (getterObject != null) {
                        // yes, then use the existing object
                        newObject = getterObject;
                    }
                }

                // Set attributes
                setAttributes((Element) node, newObject, classMapping);

                // add or set element on the given object
                addSetElement(object, newObject, node.getLocalName(), classMapping);

                // Store element
                configuredElements.put(node.getLocalName(), newObject);

                // configure sub element of the new object.
                configure(newObject, (Element) node, classMapping);
            }
        }
    }

    /**
     * Sets the attributes on the given object.
     * @param node the XML element object
     * @param object the object to configure
     * @param classMapping the mapping data
     * @throws XMLConfigurationException if the attributes cannot be set
     */
    protected void setAttributes(final Element node, final Object object, final ClassMapping classMapping)
            throws XMLConfigurationException {
        // Set attributes
        NamedNodeMap attributes = node.getAttributes();
        int attributesLength = attributes.getLength();

        // Handle special case of java.util.Properties
        if (object instanceof Properties) {
            // get name and value attributes
            String name = null;
            String value = null;
            for (int a = 0; a < attributesLength; a++) {
                Node nodeAttribute = attributes.item(a);
                if (nodeAttribute.getNamespaceURI() == null) {
                    String propertyName = nodeAttribute.getLocalName();
                    String propertyValue = nodeAttribute.getNodeValue();
                    if ("name".equals(propertyName)) {
                        name = propertyValue;
                    } else if ("value".equals(propertyName)) {
                        value = propertyValue;
                    }
                }
            }
            if (name != null && value != null) {
                ((Properties) object).put(name, value);
            } else {
                logger.warn("Properties object found but no name/value attributes were present from ''{0}''", attributes);
            }
            return;
        }


        for (int a = 0; a < attributesLength; a++) {
            Node nodeAttribute = attributes.item(a);
            if (nodeAttribute.getNamespaceURI() == null) {
                String propertyName = nodeAttribute.getLocalName();
                String propertyValue = nodeAttribute.getNodeValue();
                setAttribute(object, propertyName, propertyValue, classMapping);
            }
        }

        // this element is configured to have a value mapped to an attribute ?
        String elementAttribute = classMapping.getElementAttribute();
        if (elementAttribute != null) {
            // Get value
            Node txtNode = node.getFirstChild();
            String val = null;
            if (txtNode != null) {
                val = txtNode.getNodeValue();
            }
            setAttribute(object, elementAttribute, val, classMapping);
        }
    }

    /**
     * Gets the Mapping for the given nameSpaceURI and the given classname.
     * @param nameSpaceURI the namespace of the xml element
     * @param name the name of the class
     * @return the class mapping object
     * @throws XMLConfigurationException if no configuration is found
     */
    private ClassMapping getClassMapping(final String nameSpaceURI, final String name) throws XMLConfigurationException {
        // Get classname with the namespace and alias
        XMLMapping xmlMapping = mappings.get(nameSpaceURI);
        if (xmlMapping == null) {
            // Not yet encountered namespace. Try to get a mapping for it.
            xmlMapping = getXMLMapping(nameSpaceURI, name);
            if (xmlMapping == null) {
                // Still not found.
                throw new XMLConfigurationException("No mapping found for namespace '" + nameSpaceURI + "'.");
            }
        }

        // alias present ?
        ClassMapping classMapping = xmlMapping.getClassMapping(name);

        // no mapping
        if (classMapping == null) {
            throw new XMLConfigurationException("No class found for element '" + name + "' withing xmlmapping '" + xmlMapping
                    + "'.");
        }
        return classMapping;
    }

    /**
     * Build class instance for the given node.
     * @param node the element of the xml
     * @param classMapping the mapping between node name and class name
     * @param objectClassLoader the object class loader
     * @return the instance of the object
     * @throws XMLConfigurationException if no new instance can be done.
     */
    private Object newInstance(final Node node, final ClassMapping classMapping, final ClassLoader objectClassLoader)
            throws XMLConfigurationException {
        // get classname
        String className = classMapping.getName();

        // load class
        Class<?> clazz;
        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        try {
            clazz = classLoader.loadClass(className);
        } catch (ClassNotFoundException e) {
            try {
                clazz = objectClassLoader.loadClass(className);
            } catch (ClassNotFoundException e1) {
                throw new XMLConfigurationException("Cannot load the class '" + className + "'.", e);
            }
        }
        logger.debug("Building instance of the class {0}", className);
        // build a new instance
        try {
            return clazz.newInstance();
        } catch (InstantiationException e) {
            throw new XMLConfigurationException("Cannot build an instance of class '" + className + "' using classloader '"
                    + classLoader + "'.", e);
        } catch (IllegalAccessException e) {
            throw new XMLConfigurationException("Cannot build an instance of class '" + className + "' using classloader '"
                    + classLoader + "'.", e);
        } catch (NoClassDefFoundError e) {
            throw new XMLConfigurationException("Cannot build an instance of class '" + className + "' using classloader '"
                    + classLoader + "'.", e);
        }
    }

    /**
     * Set attribute on the given object.
     * @param configuringObject the object to configure
     * @param propertyName the name of the property (for the setter)
     * @param propertyValue the value of the property
     * @param classMapping the mapping (if there are attributes aliasing)
     * @throws XMLConfigurationException if attribute cannot be set.
     */
    private void setAttribute(final Object configuringObject, final String propertyName, final String propertyValue,
            final ClassMapping classMapping) throws XMLConfigurationException {

        // get class of the configuringObject
        Class<?> configuringClass = configuringObject.getClass();

        // Get methods
        Method[] methods = configuringClass.getMethods();

        String updatedPropertyName = propertyName;

        // alias ?
        AttributeMapping attributeMapping = classMapping.getAttributeMapping(propertyName);
        if (attributeMapping != null) {
            String name = attributeMapping.getName();
            if (name != null) {
                updatedPropertyName = name;
            }
        }

        String setterName = "set" + updatedPropertyName.substring(0, 1).toUpperCase() + updatedPropertyName.substring(1);

        Object paramObject = null;
        // Search setter
        Method foundMethod = null;
        if (methods != null) {
            for (Method m : methods) {
                // set or add ?
                if (m.getName().equals(setterName)) {
                    // only one arg ?
                    Class<?>[] parameters = m.getParameterTypes();
                    if (parameters.length == 1) {
                        foundMethod = m;
                        Class<?> parameterClass = parameters[0];

                        // Special case of reference
                        if (propertyValue.startsWith("#") && propertyValue.length() >= 2) {
                            String referenceName = propertyValue.substring(1);

                            // Try to see if there is first a matching element's name
                            paramObject = configuredElements.get(referenceName);

                            if (paramObject == null) {
                                // No matching element, try to see if there is a match in the contextual instances
                                paramObject = contextualInstances.get(referenceName);
                            }

                        // Special case of property
                        } else if (propertyValue.contains("${")
                                && propertyValue.length() >= MIN_PROPERTY_LENGTH
                                && propertyResolver != null) {
                            try {
                                paramObject =
                                    PropertyResolverHelper.resolve(propertyResolver, propertyValue);
                            } catch (IllegalArgumentException e) {
                                logger.error("Cannot resolve the property {1}", propertyValue, e);
                            }
                        } else if (String.class.equals(parameterClass)) {
                            paramObject = propertyValue;
                        } else if (URL.class.equals(parameterClass)) {
                            try {
                                paramObject = new URL(propertyValue);
                            } catch (MalformedURLException e) {
                                throw new XMLConfigurationException("Cannot convert '" + propertyValue + "' into URL"
                                                                    + " for property '" + propertyName + "'", e);
                            }
                        } else if (Byte.TYPE.equals(parameterClass) || Byte.class.equals(parameterClass)) {
                            paramObject = Byte.valueOf(propertyValue);
                        } else if (Short.TYPE.equals(parameterClass) || Short.class.equals(parameterClass)) {
                            paramObject = Short.valueOf(propertyValue);
                        } else if (Integer.TYPE.equals(parameterClass) || Integer.class.equals(parameterClass)) {
                            paramObject = Integer.valueOf(propertyValue);
                        } else if (Long.TYPE.equals(parameterClass) || Long.class.equals(parameterClass)) {
                            paramObject = Long.valueOf(propertyValue);
                        } else if (Float.TYPE.equals(parameterClass) || Float.class.equals(parameterClass)) {
                            paramObject = Float.valueOf(propertyValue);
                        } else if (Double.TYPE.equals(parameterClass) || Double.class.equals(parameterClass)) {
                            paramObject = Double.valueOf(propertyValue);
                        } else if (Boolean.TYPE.equals(parameterClass) || Boolean.class.equals(parameterClass)) {
                            paramObject = Boolean.valueOf(propertyValue);
                        }
                        break;
                    }
                }
            }
        }

        // build
        if (foundMethod != null && paramObject != null) {
            logger.debug("Calling {0} on object {1} with value {2}", setterName, configuringObject, paramObject);
            try {
                foundMethod.invoke(configuringObject, paramObject);
            } catch (IllegalArgumentException e) {
                throw new XMLConfigurationException("Cannot call method '" + foundMethod + "' on object '" + configuringObject
                        + "' with parameter '" + paramObject + "'.", e);
            } catch (IllegalAccessException e) {
                throw new XMLConfigurationException("Cannot call method '" + foundMethod + "' on object '" + configuringObject
                        + "' with parameter '" + paramObject + "'.", e);
            } catch (InvocationTargetException e) {
                throw new XMLConfigurationException("Cannot call method '" + foundMethod + "' on object '" + configuringObject
                        + "' with parameter '" + paramObject + "'.", e);
            }
        } else {
            if (foundMethod == null) {
                throw new XMLConfigurationException("No setter method (expected name: " + setterName + ") found for parameter '" + propertyName + "' on class '"
                                                    + configuringClass + "'.");
            } else if (attributeMapping == null || !attributeMapping.isOptional()) {
                throw new XMLConfigurationException("No value Object found for required parameter '" + propertyName + "' on class '"
                                                    + configuringClass + "'.");
            }
        }
    }

    /**
     * Finds the add(Class c) or set(Class c) methods.<br>
     * The parameter needs to be a single entry and not a list.
     * @param configuringObject the object on which call this setter method.
     * @param paramObject the parameter object
     * @param attributeName the name of the attribute for this object
     * @return the found method if any.
     */
    private Method findSingleSetterMethod(final Object configuringObject, final Object paramObject,
            final String attributeName) {
        Method setterMethod = null;

        // get class of the configuringObject
        Class<?> configuringClass = configuringObject.getClass();

        // Get methods of the class
        Method[] methods = configuringClass.getMethods();

        // class of the new object
        Class<?> parameterClass = paramObject.getClass();
        String className = parameterClass.getSimpleName();

        // Search setter
        if (methods != null) {
            for (Method m : methods) {
                // set or add ?
                if (m.getName().startsWith("set" + className) || m.getName().startsWith("add" + className)
                        || m.getName().equalsIgnoreCase("set" + attributeName)) {
                    // only one arg ?
                    Class<?>[] parameters = m.getParameterTypes();
                    if (parameters.length == 1) {
                        // valid parameter
                        if (parameters[0].equals(parameterClass) || parameters[0].isAssignableFrom(parameterClass)) {
                            setterMethod = m;
                            break;
                        }
                    }
                }
            }
        }
        return setterMethod;
    }

    /**
     * Finds the get(Class c) method.<br>
     * The parameter needs to be a single entry and not a list.
     * @param configuringObject the object on which call this getter method.
     * @param paramObject the parameter object
     * @return the found method if any.
     */
    private Method findSingleGetterMethod(final Object configuringObject, final Object paramObject) {
        Method getterMethod = null;

        // get class of the configuringObject
        Class<?> configuringClass = configuringObject.getClass();

        // Get methods of the class
        Method[] methods = configuringClass.getMethods();

        // class of the new object
        Class<?> parameterClass = paramObject.getClass();
        String className = parameterClass.getSimpleName();

        // Search setter
        if (methods != null) {
            for (Method m : methods) {
                // Gettermethod ?
                if (m.getName().startsWith("get" + className)) {
                    // no Arg ?
                    Class<?>[] parameters = m.getParameterTypes();
                    if (parameters.length == 0) {
                        // valid parameter and return type ?
                        Class<?> returnType = m.getReturnType();
                        if (returnType != null && returnType.equals(paramObject.getClass())) {
                            getterMethod = m;
                            break;
                        }
                    }
                }
            }
        }
        return getterMethod;
    }

    /**
     * Finds the get(List&lt;Class&gt;) and set(List&lt;Class&gt;) methods. The
     * parameter needs to be a list with generics.
     * @param configuringObject the object on which call this setter method.
     * @param paramObject the parameter object
     * @param getter the name of the getter
     * @param setter the name of the setter
     * @return the getter and setter method with List type.
     */
    private Method[] findListGetterSetterMethod(final Object configuringObject, final Object paramObject, final String getter,
            final String setter) {

        // get class of the configuringObject
        Class<?> configuringClass = configuringObject.getClass();

        // Get methods of the class
        Method[] methods = configuringClass.getMethods();

        // class of the new object
        Class<?> parameterClass = paramObject.getClass();

        // Search with class of object
        String className = parameterClass.getSimpleName();
        // Suffix name (many items)
        // (special case for y)
        String argName = null;
        if (className.endsWith("y")) {
            argName = className.substring(0, className.length() - 1) + "ies";
        } else if (className.endsWith("ss")) {
            argName = className + "es";
        } else {
            argName = className + "s";
        }

        Method[] searchMethods = searchListGetterSetterMethod(methods, argName, parameterClass, getter, setter);

        return searchMethods;
    }

    /**
     * Search getter and setter method in the given set of methods. The
     * getter/setter are used for a List parameter/return type.
     * @param methods the set of methods to search in
     * @param argName the name of the argument
     * @param parameterClass the parameter class (for the generics on the List)
     * @param getter the name of the getter
     * @param setter the name of the setter
     * @return an array of methods, array[0] = getter, array[1] = setter.
     */
    private Method[] searchListGetterSetterMethod(final Method[] methods, final String argName, final Class<?> parameterClass,
            final String getter, final String setter) {
        Method[] returnedMethods = null;

        Method getterMethod = null;
        Method setterMethod = null;
        // Search getter and setter
        if (methods != null) {
            for (Method m : methods) {
                // setter
                if (m.getName().startsWith("set" + argName) || m.getName().equals(setter)) {
                    // only one arg ?
                    Class<?>[] parameters = m.getParameterTypes();

                    // setter
                    if (parameters.length == 1) {
                        // ensure it is a List type
                        if (parameters[0].equals(List.class)) {
                            Type[] types = m.getGenericParameterTypes();
                            // look only the first type if any
                            if (types.length == 1) {
                                if (types[0] instanceof ParameterizedType) {
                                    ParameterizedType parameterizedType = (ParameterizedType) types[0];
                                    Type[] typeArguments = parameterizedType.getActualTypeArguments();
                                    if (typeArguments.length == 1) {
                                        if (typeArguments[0].equals(parameterClass)) {
                                            // found it !
                                            setterMethod = m;
                                        }
                                    }

                                }
                            }
                        }
                    }
                } else if (m.getName().startsWith("get" + argName) || m.getName().equals(getter)) {
                    Class<?> returnTypeClass = m.getReturnType();
                    if (returnTypeClass != null && returnTypeClass.equals(List.class)) {

                        // no parameter
                        Class<?>[] parameters = m.getParameterTypes();
                        if (parameters.length == 0) {
                            Type type = m.getGenericReturnType();
                            if (type instanceof ParameterizedType) {
                                ParameterizedType parameterizedType = (ParameterizedType) type;
                                Type[] typeArguments = parameterizedType.getActualTypeArguments();
                                if (typeArguments.length == 1) {
                                    if (typeArguments[0].equals(parameterClass)) {
                                        // found it !
                                        getterMethod = m;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        // try with interfaces of the current class
        Class<?>[] interfaces = parameterClass.getInterfaces();
        if (interfaces != null) {
            for (Class<?> itf : interfaces) {
                String itfName = itf.getSimpleName();
                String argItfName = itfName + "s";
                returnedMethods = searchListGetterSetterMethod(methods, argItfName, itf, getter, setter);
                if (returnedMethods != null) {
                    break;
                }
            }
        }

        if (getterMethod != null && setterMethod != null) {
            returnedMethods = new Method[2];
            returnedMethods[0] = getterMethod;
            returnedMethods[1] = setterMethod;
        }
        return returnedMethods;
    }


    /**
     * Add or set element on the given object.
     * @param configuringObject the object to configure
     * @param newObject the object to add/set on the configuringObject object
     * @param attributeName the name of the attribute for this object
     * @param classMapping the data about all attributes
     * @throws XMLConfigurationException if newObject cannot be set.
     */
    @SuppressWarnings("unchecked")
    private void addSetElement(final Object configuringObject, final Object newObject, final String attributeName,
            final ClassMapping classMapping) throws XMLConfigurationException {

        // Getter and setter set ?
        String getter = null;
        String setter = null;
        boolean isListAttribute = false;

        // Read properties if any
        AttributeMapping attributeMapping = classMapping.getAttributeMapping(attributeName);
        if (attributeMapping != null) {
            getter = attributeMapping.getGetter();
            setter = attributeMapping.getSetter();
            isListAttribute = attributeMapping.isListElement();
        }


        // Single setter method object
        Method foundMethod = null;

        // Search a single setter if not a list
        if (!isListAttribute) {
            foundMethod =  findSingleSetterMethod(configuringObject, newObject, attributeName);
        }

        // single setter ?
        if (foundMethod != null) {
            logger.debug("Calling method {0} on object {1} with value {2}", foundMethod, configuringObject, newObject);
            try {
                foundMethod.invoke(configuringObject, newObject);
            } catch (IllegalArgumentException e) {
                throw new XMLConfigurationException("Unable to invoke the method '" + foundMethod + "' on '" + configuringObject
                        + "' with arg '" + newObject + "'.", e);
            } catch (IllegalAccessException e) {
                throw new XMLConfigurationException("Unable to invoke the method '" + foundMethod + "' on '" + configuringObject
                        + "' with arg '" + newObject + "'.", e);
            } catch (InvocationTargetException e) {
                throw new XMLConfigurationException("Unable to invoke the method '" + foundMethod + "' on '" + configuringObject
                        + "' with arg '" + newObject + "'.", e);
            }
        } else {
            // try a list getter/setter
            Method[] getterAndSetter = findListGetterSetterMethod(configuringObject, newObject, getter, setter);
            if (getterAndSetter != null) {
                Method getterMethod = getterAndSetter[0];
                Method setterMethod = getterAndSetter[1];

                // Get current list
                List currentList = null;
                try {
                    logger.debug("Calling method {0} on object {1}", getterMethod, configuringObject);
                    currentList = (List) getterMethod.invoke(configuringObject);
                } catch (IllegalArgumentException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + getterMethod + "' on '"
                            + configuringObject + "'.", e);
                } catch (IllegalAccessException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + getterMethod + "' on '"
                            + configuringObject + "'.", e);
                } catch (InvocationTargetException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + getterMethod + "' on '"
                            + configuringObject + "'.", e);
                }

                // Null list, create one
                if (currentList == null) {
                    currentList = new ArrayList();
                }

                // add element
                logger.debug("Adding element  {0} to the list {1}", newObject, currentList);
                currentList.add(newObject);

                // set list
                try {
                    logger.debug("Calling method {0} on object {1} with value {2}", setterMethod, configuringObject, currentList);
                    setterMethod.invoke(configuringObject, currentList);
                } catch (IllegalArgumentException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + setterMethod + "' on '"
                            + configuringObject + "'.", e);
                } catch (IllegalAccessException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + setterMethod + "' on '"
                            + configuringObject + "'.", e);
                } catch (InvocationTargetException e) {
                    throw new XMLConfigurationException("Unable to invoke the method '" + setterMethod + "' on '"
                            + configuringObject + "'.", e);
                }

            } else {
                throw new XMLConfigurationException("No setter method found for parameter '" + newObject.getClass()
                        + "' on class '" + configuringObject.getClass() + "'.");
            }
        }
    }

    /**
     * Gets the XML mapping for the given namespace and the node name.
     * @param nameSpace current XML namespace.
     * @param optionalName the name to use if default name is not available
     * @return the XML mapping object.
     * @throws XMLConfigurationException if xml mapping is not found.
     */
    private XMLMapping getXMLMapping(final String nameSpace, final String optionalName) throws XMLConfigurationException {
        if (nameSpace == null || "".equals(nameSpace)) {
            throw new XMLConfigurationException("Empty default namespace not supported.");
        }

        // already one ?
        if (mappings.containsKey(nameSpace)) {
            return mappings.get(nameSpace);
        }

        XMLMapping xmlMapping = null;
        // With the namespace, get the package name for searching the
        // <mappingFilename> file or <optionalName>-mapping.xml
        if (nameSpace.startsWith(HTTP_NAMESPACE)) {
            // get package
            String packageName = nameSpace.substring(HTTP_NAMESPACE.length());
            // Compute node name and suffix
            String resourceName = packageName.replaceAll("\\.", "/") + "/" + mappingFilename;

            URL mappingURL = findResource(resourceName);

            // try with the name of the element if not found
            if (mappingURL == null && optionalName != null) {
                resourceName = packageName.replaceAll("\\.", "/") + "/" + optionalName + MAPPING_SUFFIX;
                mappingURL = findResource(resourceName);
            }

            // Element has been found, create mapping from this object
            if (mappingURL != null) {
                xmlMapping = createXMLMapping(mappingURL);
            }
        } else {
            logger.warn("Namespace found ''{0}}'' but this is not an HTTP namespace", nameSpace);
        }
        // even null namespace will have a default xml mapping
        mappings.put(nameSpace, xmlMapping);
        return xmlMapping;

    }

    /**
     * Find a given resource name in different ClassLoaders (TCCL and the
     * XMLConfiguration's own ClassLoader).
     * @param resourceName name of the resource to be found.
     * @return an URL pointing to the resource or null if nothing was found
     */
    private URL findResource(final String resourceName) {

        // First try to lookup the resource from the TCCL
        URL mappingURL = Thread.currentThread().getContextClassLoader().getResource(resourceName);

        // If not found use this class' ClassLoader
        if (mappingURL == null) {
            ClassLoader loader = XMLConfiguration.class.getClassLoader();
            mappingURL = loader.getResource(resourceName);
        }

        // Be careful, the return value can still be null at this point !
        return mappingURL;
    }

    /**
     * Create the XML mapping object for the given namespace with the given XML
     * mapping file.
     * @param mappingURL the URL to the XML mapping file
     * @return an instance of XML mapping
     * @throws XMLConfigurationException if no XML mapping have been created.
     */
    private XMLMapping createXMLMapping(final URL mappingURL) throws XMLConfigurationException {
        XMLMappingBuilder xmlMappingBuilder = new XMLMappingBuilder(mappingURL);
        xmlMappingBuilder.build();
        return xmlMappingBuilder.getXmlMapping();
    }

    /**
     * @return the property resolver
     */
    public IPropertyResolver getPropertyResolver() {
        return propertyResolver;
    }

    /**
     * Set the property resolver.
     * @param propertyResolver the property resolver
     */
    public void setPropertyResolver(final IPropertyResolver propertyResolver) {
        this.propertyResolver = propertyResolver;
    }

    /**
     * @return the contextual instances Map
     */
    public Map<String, Object> getContextualInstances() {
        return contextualInstances;
    }

    /**
     * Set the contextual instances Map.
     * @param contextualInstances the contextual instances Map
     */
    public void setContextualInstances(final Map<String, Object> contextualInstances) {
        this.contextualInstances = contextualInstances;
    }

}
