/**
 * Copyright 2006-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.xmlconfig.mapping;

import java.net.URL;
import java.util.List;

import org.ow2.util.xml.DocumentParser;
import org.ow2.util.xml.DocumentParserException;
import org.ow2.util.xml.EmptyEntityResolver;
import org.ow2.util.xml.XMLUtils;
import org.ow2.util.xmlconfig.XMLConfigurationException;
import org.ow2.util.xmlconfig.ns.MappingResolver;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Allows to build the Mapping object for a set of classes.
 * @author Florent Benoit
 */
public class XMLMappingBuilder {

    /**
     * Name of the &lt;package&gt; element.
     */
    private static final String PACKAGE_ELEMENT = "package";

    /**
     * Name of the &lt;class&gt; element.
     */
    private static final String CLASS_ELEMENT = "class";

    /**
     * Name of the &lt;attribute&gt; element.
     */
    private static final String ATTRIBUTE_ELEMENT = "attribute";

    /**
     * URL that reference the mapping file.
     */
    private URL mappingURL = null;

    /**
     * The mapping used for the given namespace.
     */
    private XMLMapping xmlMapping = null;

    /**
     * Validating ?
     */
    private boolean validating = !Boolean.getBoolean("xmlconfig.validating.disable");

    /**
     * Builds a new mapping builder.
     * @param mappingURL the given url that reference the mapping file.
     */
    public XMLMappingBuilder(final URL mappingURL) {
        this.mappingURL = mappingURL;
        this.xmlMapping = new XMLMapping();
    }

    /**
     * Builds a new mapping builder. (only for backward compliance)
     * @param nameSpace the namespace of the mapping
     * @param mappingURL the given url that reference the mapping file.
     * @deprecated
     */
    @Deprecated
    public XMLMappingBuilder(final String nameSpace, final URL mappingURL) {
        this(mappingURL);
    }

    /**
     * Build the XMLMapping object by analyzing the mapping file.
     * @throws XMLConfigurationException if there is a failure when analyzing
     *         the XML file.
     */
    public void build() throws XMLConfigurationException {

        // Parse the XML file and build all configuration process.
        Document xmlMappingConfigurationDocument = null;
        try {
            if (validating) {
                xmlMappingConfigurationDocument = DocumentParser.getDocument(mappingURL, true, new MappingResolver());
            } else {
                xmlMappingConfigurationDocument = DocumentParser.getDocument(mappingURL, false, new EmptyEntityResolver());
            }
        } catch (DocumentParserException e) {
            throw new XMLConfigurationException("Cannot get a document on the given url '" + mappingURL + "'.", e);
        }

        // Get the root element
        Element rootMappingElement = xmlMappingConfigurationDocument.getDocumentElement();

        // <package> element ?
        NodeList packageList = rootMappingElement.getElementsByTagName(PACKAGE_ELEMENT);
        for (int i = 0; i < packageList.getLength(); i++) {
            Element packageElement = (Element) packageList.item(i);

            // get name
            String packageName = XMLUtils.getAttributeValue(packageElement, "name");

            // Create class mapping
            NodeList classList = packageElement.getElementsByTagName(CLASS_ELEMENT);
            addClassMapping(classList, packageName + ".", true);
        }

        // Now analyze single <class> element
        NodeList classList = rootMappingElement.getElementsByTagName(CLASS_ELEMENT);
        addClassMapping(classList, "", false);
    }

    /**
     * Add the mapping for the given list of class elements.
     * @param classList the list of elements
     * @param packageName the name of the package to use as prefix.
     * @param packageParent if true, package as parent node is accepted, else it
     *        is denied.
     */
    private void addClassMapping(final NodeList classList, final String packageName, final boolean packageParent) {

        // class elements ?
        for (int c = 0; c < classList.getLength(); c++) {
            Element classElement = (Element) classList.item(c);

            // if package element is parent but packageElement boolean is false,
            // stop
            if (classElement.getParentNode().getNodeName().equals(PACKAGE_ELEMENT) && !packageParent) {
                continue;
            }

            // Build mapping object
            ClassMapping classMapping = new ClassMapping();

            // class name
            String name = XMLUtils.getAttributeValue(classElement, "name");
            String className = packageName + name;
            classMapping.setName(className);

            // alias
            String alias = XMLUtils.getAttributeValue(classElement, "alias");
            if (alias != null) {
                classMapping.getAliases().add(alias);
            }

            // aliases ?
            NodeList aliasesList = classElement.getElementsByTagName("aliases");
            if (aliasesList.getLength() > 0) {
                List<String> aliases = XMLUtils.getStringListValueElement((Element) aliasesList.item(0), "alias");
                classMapping.getAliases().addAll(aliases);
            }

            // Mapping of the content of an element to an attribute
            String elementAttribute = XMLUtils.getAttributeValue(classElement, "element-attribute");
            classMapping.setElementAttribute(elementAttribute);

            // Attributes ?
            NodeList attributeList = classElement.getElementsByTagName(ATTRIBUTE_ELEMENT);
            for (int a = 0; a < attributeList.getLength(); a++) {
                Element attributeElement = (Element) attributeList.item(a);
                // Build mapping object
                AttributeMapping attributeMapping = new AttributeMapping();

                // name
                String attributeName = XMLUtils.getAttributeValue(attributeElement, "name");
                attributeMapping.setName(attributeName);

                // alias
                String attributeAlias = XMLUtils.getAttributeValue(attributeElement, "alias");
                if (attributeAlias != null) {
                    attributeMapping.getAliases().add(attributeAlias);
                }

                // aliases ?
                NodeList attributesAliasesList = attributeElement.getElementsByTagName("aliases");
                if (attributesAliasesList.getLength() > 0) {
                    List<String> aliases = XMLUtils.getStringListValueElement((Element) aliasesList.item(0), "alias");
                    attributeMapping.getAliases().addAll(aliases);
                }

                // is Element ?
                String isElementName = XMLUtils.getAttributeValue(attributeElement, "element");
                if (Boolean.parseBoolean(isElementName)) {
                    attributeMapping.setElement();
                }

                // is List Element ?
                String isListElementName = XMLUtils.getAttributeValue(attributeElement, "isList");
                if (Boolean.parseBoolean(isListElementName)) {
                    attributeMapping.setListElement();
                }

                // is optional ?
                String isOptional = XMLUtils.getAttributeValue(attributeElement, "optional");
                if (Boolean.parseBoolean(isOptional)) {
                    attributeMapping.setOptional();
                }

                // getter
                String getter = XMLUtils.getAttributeValue(attributeElement, "getter");
                attributeMapping.setGetter(getter);

                // setter
                String setter = XMLUtils.getAttributeValue(attributeElement, "setter");
                attributeMapping.setSetter(setter);

                // add attribute
                classMapping.addAttributeMapping(attributeMapping);
            }

            xmlMapping.addClassMapping(classMapping);
        }
    }

    /**
     * Gets the XML mapping object.
     * @return the mapping object between classname/alias and their mapping
     *         description.
     */
    public XMLMapping getXmlMapping() {
        return xmlMapping;
    }
}
